import itertools
from collections import Counter
from typing import Any, Dict, Iterable, Iterator, List, Optional, Sequence, Sized, Type

from d8s_dicts import dict_sort_by_values, dict_values

# TODO: consider applying @decorators.listify_first_arg argument to all/most functions in this module


def iterable_sort_by_length(iterable: Iterable[Any], **kwargs) -> List[Any]:
    """."""
    sorted_list = sorted(iterable, key=lambda x: len(x), **kwargs)  # pylint: disable=W0108
    return sorted_list


def longest(iterable: Iterable[Any]) -> Any:
    """."""
    longest_item = max(iterable, key=len)
    return longest_item


def shortest(iterable: Iterable[Any]) -> Any:
    """."""
    shortest_item = min(iterable, key=len)
    return shortest_item


def flatten(iterable: Iterable[Any], level: int = None, **kwargs) -> Iterator[Any]:
    """Flatten all items in the iterable so that they are all items in the same list."""
    import more_itertools

    return more_itertools.collapse(iterable, levels=level, **kwargs)


def has_index(iterable: Sized, index: int) -> bool:
    """."""
    # TODO: would it be faster to simply try to get the item at index and handle exceptions?
    index_int = int(index)

    return 0 <= index_int <= len(iterable) - 1


def types(iterable: Iterable[Any]) -> Iterator[Type]:
    """Return a set containing the types of all items in the list_arg."""
    return map(type, iterable)


def iterable_item_of_types(iterable: Iterable[Any], item_types: Iterable[type]) -> bool:
    """Return True if the iterable has any item that is of the item_types. Otherwise, return False."""
    for i in iterable:
        if type(i) in item_types:
            return True
    return False


def iterable_all_items_of_types(iterable: Iterable[Any], item_types: Iterable[type]) -> bool:
    """Return True if all items in the iterable are of a type given in item_types. Otherwise, return False."""
    for i in iterable:
        if type(i) not in item_types:
            return False
    return True


def iterable_has_all_items_of_type(iterable: Iterable[Any], type_arg: type) -> bool:
    """Return whether or not all iterable in iterable are of the type specified by the type_arg."""
    for i in iterable:
        if type(i) != type_arg:
            return False
    return True


def deduplicate(iterable: Iterable[Any]) -> Iterator[Any]:
    """Deduplicate the iterable."""
    og_iterable, temp_iterable = itertools.tee(iterable)

    if iterable_item_of_types(temp_iterable, (dict, list)):
        deduplicated_list = []
        for i in og_iterable:
            if i not in deduplicated_list:
                deduplicated_list.append(i)
                yield i
    else:
        yield from list(set(og_iterable))


def cycle(iterable: Iterable[Any], length: Optional[int] = None) -> Iterator[Any]:
    """Cycle through the iterable as much as needed."""
    full_cycle = itertools.cycle(iterable)

    if length:
        for index, item in enumerate(full_cycle):
            yield item
            if index == length - 1:
                break
    else:
        return full_cycle


def truthy_items(iterable: Iterable[Any]) -> Iterator[Any]:
    """Return an iterable with only elements of the given iterable which evaluate to True.

    (see https://docs.python.org/3.9/library/stdtypes.html#truth-value-testing)
    """
    return filter(lambda x: x, iterable)


def nontruthy_items(iterable: Iterable[Any]) -> Iterator[Any]:
    """Return an iterable with only elements of the given iterable which evaluate to False.

    (see https://docs.python.org/3.9/library/stdtypes.html#truth-value-testing)
    """
    return filter(lambda x: not x, iterable)


def iterable_has_single_item(iterable: Iterable[Any]) -> bool:
    """Return whether the iterable has a single item in it."""
    iterable = deduplicate(iterable)
    result = len(tuple(iterable)) == 1
    return result


def iterables_are_same_length(a: Sized, b: Sized, *args: Sized, debug_failure: bool = False) -> bool:
    """Return whether or not the given iterables are the same lengths."""
    consolidated_list = [a, b, *args]
    lengths_1, lengths_2 = itertools.tee(map(len, consolidated_list))

    result = iterable_has_single_item(lengths_1)

    if debug_failure and not result:
        list_length_breakdown = iterable_count(lengths_2)
        minority_list_count = min(dict_values(list_length_breakdown))
        for index, arg in enumerate(consolidated_list):
            if list_length_breakdown[len(arg)] == minority_list_count:
                print(f'Argument {index} is not the same length as the majority of the arguments')

    return result


def iterables_have_same_items(a: Sequence, b: Sequence, *args: Sequence) -> bool:  # noqa: CCR001
    """Return whether iterables have identical items (considering both identity and count)."""
    first_list = a
    remaining_lists = [b, *args]

    if iterables_are_same_length(a, *remaining_lists):
        for item in first_list:
            first_list_count = first_list.count(item)
            item_counts = [list_.count(item) for list_ in remaining_lists]
            same_count = item_counts[0] == first_list_count
            if not iterable_has_single_item(item_counts) or not same_count:
                return False
    else:
        return False
    return True


def run_length_encoding(iterable: Iterable[Any]) -> Iterator[str]:
    """Perform run-length encoding on the given array.

    See https://en.wikipedia.org/wiki/Run-length_encoding for more details.
    """
    run_length_encodings = (f'{len(tuple(g))}{k}' for k, g in itertools.groupby(iterable))
    return run_length_encodings


def iterable_count(iterable: Iterable[Any]) -> Dict[Any, int]:
    """Count each item in the iterable."""
    counter: Counter = Counter()
    counter.update(iterable)
    count = dict_sort_by_values(counter)
    return count


def iterable_item_index(iterable: Sequence, item: Any) -> int:
    """Find the given item in the iterable. Return -1 if the item is not found."""
    try:
        return iterable.index(item)
    except ValueError:
        return -1


def iterable_item_indexes(iterable: Iterable[Any], item: Any) -> Iterator[int]:
    """Find the given item in the iterable. Return -1 if the item is not found."""
    indexes = (index for index, value in enumerate(iterable) if value == item)
    return indexes


def duplicates(iterable: Sequence) -> Iterator[Sequence]:
    """Find duplicates in the given iterable."""
    for item in iterable:
        if iterable.count(item) > 1:
            yield item


def iterable_has_mixed_types(iterable: Iterable[Any]) -> bool:
    """Return whether or not the iterable has items with two or more types."""
    return len(tuple(deduplicate(types(iterable)))) >= 2


def iterable_has_single_type(iterable: Iterable[Any]) -> bool:
    """Return whether or not the iterable has items of only one type."""
    return len(tuple(deduplicate(types(iterable)))) == 1


def iterable_replace(iterable: Iterable[Any], old_value: Any, new_value: Any) -> Iterator[Any]:
    """Replace all instances of the old_value with the new_value in the given iterable."""
    for value in iterable:
        if value == old_value:
            yield new_value
        else:
            yield value


# def list_entropy(list_arg):
#     """Find the entropy of the items in the given list."""
#     import math
#     from nlp import frequencyDistribution

#     freqdist = frequencyDistribution(iterable)
#     probs = [freqdist.freq(l) for l in freqdist]
#     return -sum(p * math.log(p, 2) for p in probs)



from collections import OrderedDict

import pytest

from d8s_lists import (
    cycle,
    deduplicate,
    duplicates,
    flatten,
    has_index,
    iterable_all_items_of_types,
    iterable_count,
    iterable_has_all_items_of_type,
    iterable_has_mixed_types,
    iterable_has_single_item,
    iterable_has_single_type,
    iterable_item_index,
    iterable_item_indexes,
    iterable_item_of_types,
    iterable_replace,
    iterable_sort_by_length,
    iterables_are_same_length,
    iterables_have_same_items,
    longest,
    nontruthy_items,
    run_length_encoding,
    shortest,
    truthy_items,
    types,
)

# from pyannotate_runtime import collect_types

# collect_types.init_types_collection()
# collect_types.start()


@pytest.mark.parametrize("test_input,result_contents", [((None, False, [], 0), ()), ([1, 0, True, False], (1, True))])
def test_truthy_items_1(test_input, result_contents):
    result = tuple(truthy_items(test_input))
    assert result == result_contents


@pytest.mark.parametrize(
    "test_input,result_contents", [((None, False, [], 0), (None, False, [], 0)), ([1, 0, True, False], (0, False))]
)
def test_nontruthy_items_1(test_input, result_contents):
    result = tuple(nontruthy_items(test_input))
    assert result == result_contents


def test_iterable_has_all_items_of_type_1():
    assert not iterable_has_all_items_of_type([1, 'a'], int)
    assert iterable_has_all_items_of_type([1, 2, 3, 4, 5], int)
    assert iterable_has_all_items_of_type([1, 2], int)
    assert not iterable_has_all_items_of_type([1, {2}], int)


def test_iterable_all_items_of_types_1():
    assert not iterable_all_items_of_types([1, 'a'], (int,))
    assert iterable_all_items_of_types([1, 'a'], (int, str))
    assert iterable_all_items_of_types([1, 'a', 3, 4], (int, str))
    assert not iterable_all_items_of_types([1, 'a', 3, {4}], (int, str))


def test_iterable_has_mixed_types_1():
    assert iterable_has_mixed_types([0, 'a'])
    assert iterable_has_mixed_types([0, 'a'])
    assert not iterable_has_mixed_types([0, 1])
    assert not iterable_has_mixed_types([{}, {}])


def test_iterable_has_single_type_1():
    assert not iterable_has_single_type([0, 'a'])
    assert not iterable_has_single_type([0, 'a'])
    assert iterable_has_single_type([0, 1])
    assert iterable_has_single_type([{}, {}])


def test_iterable_has_single_item_1():
    assert not iterable_has_single_item([])
    assert iterable_has_single_item([0])
    assert iterable_has_single_item([0, 0])
    assert not iterable_has_single_item([0, 1])


def test_iterables_have_same_items_docs_1():
    assert iterables_have_same_items([1], [1])
    assert not iterables_have_same_items([1], [2])
    assert not iterables_have_same_items([1], [1, 1])

    assert iterables_have_same_items([2, 1], [1, 2])
    assert not iterables_have_same_items([1], [2, 1])
    assert not iterables_have_same_items([2, 1], [1])
    assert not iterables_have_same_items([2, 2], [1, 2])

    assert iterables_have_same_items([1, 2, 3], [3, 1, 2])
    assert not iterables_have_same_items([1, 2, 3, 4], [3, 1, 2])
    assert not iterables_have_same_items([1, 2, 3], [3, 1, 2, 4])
    assert not iterables_have_same_items([1, 3, 3], [3, 3, 2])


def test_iterables_have_same_items_docs_2():
    assert iterables_have_same_items([1, 2, 3], [3, 2, 1])
    assert iterables_have_same_items([1, 2, 3], [3, 1, 2])
    assert not iterables_have_same_items([1, 2, 3, 3], [1, 2, 2, 3])
    assert not iterables_have_same_items([1], [1, 2])


def test_iterable_item_of_types_1():
    assert iterable_item_of_types([{'a': 1}, 1, 2, 3], (dict,))
    assert not iterable_item_of_types([1, 2, 3], (dict,))


def test_longest_1():
    l = ['a', 'aa', 'aaa']
    result = longest(l)
    assert result == 'aaa'


def test_shortest_1():
    l = ['a', 'aa', 'aaa']
    result = shortest(l)
    assert result == 'a'


def test_iterable_sort_by_length_1():
    l = ['a', 'aa', 'aaa']
    result = iterable_sort_by_length(l)
    assert result == ['a', 'aa', 'aaa']

    l = ['a', 'aa', 'aaa']
    result = iterable_sort_by_length(l, reverse=True)
    assert result == ['aaa', 'aa', 'a']

    l = ('a', 'aa', 'aaa')
    result = iterable_sort_by_length(l, reverse=True)
    assert result == ['aaa', 'aa', 'a']


def test_iterable_item_indexes_1():
    assert tuple(iterable_item_indexes([1, 2, 1, 2], 1)) == (0, 2)
    assert tuple(iterable_item_indexes([1, 2, 1, 2], 2)) == (1, 3)
    assert tuple(iterable_item_indexes([1, 2], 3)) == ()


def test_flatten_1():
    assert tuple(flatten([1, 2, 3])) == (1, 2, 3)
    assert tuple(flatten([1, [2], 3])) == (1, 2, 3)
    assert tuple(flatten([1, [2, 3]])) == (1, 2, 3)
    assert tuple(flatten([(1, 2), ([3, 4], [[5], [6]])])) == (1, 2, 3, 4, 5, 6)
    assert tuple(flatten([(1, 2), ([3, 4], [[5], [6]])], level=1)) == (1, 2, [3, 4], [[5], [6]])


def test_has_index_1():
    assert not has_index(['a', 'b', 'c'], -1)
    assert has_index(['a', 'b', 'c'], 0)
    assert has_index(['a', 'b', 'c'], 1)
    assert has_index(['a', 'b', 'c'], 2)
    assert not has_index(['a', 'b', 'c'], 3)
    assert not has_index(['a', 'b', 'c'], 4)
    assert not has_index(['a', 'b', 'c'], 4000)


def test_iterables_are_same_length_1():
    l1 = ['a']
    l2 = ['b']
    l3 = ['a']
    l4 = ['a', 'b']
    l5 = []

    assert iterables_are_same_length(l1, l2)
    assert iterables_are_same_length(l1, l2, l3)
    assert not iterables_are_same_length(l1, l2, l3, l4)
    assert not iterables_are_same_length(l1, l2, l3, l4, l5)
    assert iterables_are_same_length(l1, l3)
    assert not iterables_are_same_length(l1, l4)
    assert not iterables_are_same_length(l4, l5)
    assert not iterables_are_same_length(l1, l5)

    assert not iterables_are_same_length(l1, l4, debug_failure=True)
    assert not iterables_are_same_length(l1, l2, l3, l4, l5, debug_failure=True)


def test_types_1():
    assert tuple(types([1, 2, 3])) == (int, int, int)
    assert tuple(types(['a', 'b', 'c'])) == (str, str, str)


def test_iterable_replace_docs_1():
    old_list = [1, 2, 3]
    new_list = tuple(iterable_replace(old_list, 1, 4))
    assert new_list == (4, 2, 3)
    # test immutability
    assert old_list == [1, 2, 3]

    old_list = [1, 2, 3]
    new_list = tuple(iterable_replace(old_list, 5, 3))
    assert new_list == (1, 2, 3)
    # test immutability
    assert old_list == [1, 2, 3]

    old_list = [1, 1, 1]
    new_list = tuple(iterable_replace(old_list, 1, 0))
    assert new_list == (0, 0, 0)
    # test immutability
    assert old_list == [1, 1, 1]

    old_list = [1, 2, 3]
    new_list = tuple(iterable_replace(old_list, 1, 4))
    assert new_list == (4, 2, 3)
    # test immutability
    assert old_list == [1, 2, 3]

    old_list = [[1], [2, 2], [3, 3, 3]]
    new_list = tuple(iterable_replace(old_list, [1], [4, 4, 4, 4]))
    assert new_list == ([4, 4, 4, 4], [2, 2], [3, 3, 3])
    # test immutability
    assert old_list == [[1], [2, 2], [3, 3, 3]]

    old_list = [[{'a': 1}], [{'b': 2}], [{'c': 3}]]
    new_list = tuple(iterable_replace(old_list, [{'a': 1}], [{'e': 1}]))
    assert new_list == ([{'e': 1}], [{'b': 2}], [{'c': 3}])
    # test immutability
    assert old_list == [[{'a': 1}], [{'b': 2}], [{'c': 3}]]


def test_deduplicate_1():
    assert tuple(deduplicate([1, 2, 3, 3, 3, 4, 2])) == (1, 2, 3, 4)


def test_deduplicate_dicts():
    assert tuple(deduplicate([{'a': 1}, 1, {'b': 2}])) == ({'a': 1}, 1, {'b': 2})
    assert tuple(deduplicate([{'a': 1}, {'a': 1}, {'b': 2}])) == ({'a': 1}, {'b': 2})


def test_duplicates():
    assert tuple(duplicates([1, 2, 3, 3, 2])) == (2, 3, 3, 2)
    assert tuple(duplicates([1, 2])) == ()
    assert tuple(duplicates([1, 2, 2])) == (2, 2)

    assert tuple(duplicates([1, 2, 3, 3, 2])) == (2, 3, 3, 2)
    assert tuple(duplicates([1, 2, 2, 2, 2, 3])) == (2, 2, 2, 2)
    assert tuple(duplicates([1, 2, 1, 2])) == (1, 2, 1, 2)


def test_run_length_encoding_1():
    assert tuple(run_length_encoding('foobar')) == (
        '1f',
        '2o',
        '1b',
        '1a',
        '1r',
    )
    assert tuple(run_length_encoding(list('foobar'))) == (
        '1f',
        '2o',
        '1b',
        '1a',
        '1r',
    )

    assert tuple(run_length_encoding(['a'])) == ('1a',)
    assert tuple(run_length_encoding(['a', 'a'])) == ('2a',)
    assert tuple(run_length_encoding(['a', 'b'])) == (
        '1a',
        '1b',
    )
    assert tuple(run_length_encoding(['a', 'a', 'b'])) == (
        '2a',
        '1b',
    )
    assert tuple(run_length_encoding(['a', 'b', 'b'])) == (
        '1a',
        '2b',
    )
    assert tuple(run_length_encoding(['a', 'b', 'a'])) == (
        '1a',
        '1b',
        '1a',
    )


def test_run_length_encoding_2():
    assert tuple(run_length_encoding(['aa'])) == ('1aa',)
    assert tuple(run_length_encoding(['aa', 'aa'])) == ('2aa',)
    assert tuple(run_length_encoding(['aa', 'bbb'])) == (
        '1aa',
        '1bbb',
    )
    assert tuple(run_length_encoding(['aa', 'aa', 'bbb'])) == (
        '2aa',
        '1bbb',
    )
    assert tuple(run_length_encoding(['aa', 'bbb', 'bbb'])) == (
        '1aa',
        '2bbb',
    )
    assert tuple(run_length_encoding(['aa', 'bbb', 'aa'])) == (
        '1aa',
        '1bbb',
        '1aa',
    )
    assert tuple(run_length_encoding('WWWWWWWWWWWWBWWWWWWWWWWWWBBBWWWWWWWWWWWWWWWWWWWWWWWWBWWWWWWWWWWWWWW')) == (
        '12W',
        '1B',
        '12W',
        '3B',
        '24W',
        '1B',
        '14W',
    )


def test_multiple_duplicates():
    assert tuple(duplicates([1, 2, 3, 3, 2, 1, 3, 4, 5])) == (1, 2, 3, 3, 2, 1, 3)


def test_iterable_count_1():
    assert iterable_count([1, 2, 3, 2, 3]) == OrderedDict([(1, 1), (2, 2), (3, 2)])
    assert iterable_count(['bob', 'bob', 'frank', 'bob', 'john', 'frank', 'tim', 'tim']) == OrderedDict(
        [('john', 1), ('frank', 2), ('tim', 2), ('bob', 3)]
    )


def test_iterable_item_index_1():
    assert iterable_item_index(['a', 'b'], 'a') == 0
    assert iterable_item_index(['a', 'b'], 'b') == 1
    assert iterable_item_index(['a', 'b'], 'c') == -1


def test_cycle_1():
    l = tuple(cycle([1, 2, 3], length=42))
    assert len(l) == 42
    assert l[0] == 1
    assert l[1] == 2
    assert l[2] == 3
    assert l[3] == 1
    # collect_types.stop()
    # collect_types.dump_stats('type_info.json')


# TODO: write test for `remove_nonexistent_items` that test the following cases:
# ('', [], 0, False, None)
